/*

GMP_ext.h -- extension to the GMP float functions for the iRRAM library
 
Copyright (C) 2002-2006 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/

#ifndef GMP_ext
#define GMP_ext

#include <stdlib.h>
#include <stdio.h>

#include "GMP_interface.h"

#define iRRAM_PACKAGE PACKAGE 
#define iRRAM_VERSION VERSION
#undef PACKAGE
#undef VERSION
#include "gmp-impl.h"
#undef PACKAGE
#undef VERSION
#define PACKAGE iRRAM_PACKAGE 
#define VERSION iRRAM_VERSION
#undef iRRAM_PACKAGE
#undef iRRAM_VERSION

#include "longlong.h"

#define ABS(x) ((x) >= 0 ? (x) : -(x))
#define MAX_OF(h,i) ((h) >= (i) ? (h) : (i))



#ifdef __cplusplus
extern "C" {
#endif


#define GMP_min -1000000000
#define GMP_max 1000000000



#define GMP_PREC(x) ((unsigned int) (x)->_mp_prec * mp_bits_per_limb - mp_bits_per_limb)
//#define GMP_PREC(x) mpf_get_prec(x)

//#define GMP_SET_PREC(x,n) { if ( GMP_PREC(x)!= (n) ) {ext_gmp_set_prec(x,n);}}
#define GMP_SET_PREC(x,n) {{ext_gmp_set_prec(x,n);}}
//#define GMP_SET_PREC(x,n) { if ( GMP_PREC(x)!= (n) ) {mpf_set_prec(x,n);}}



int ext_gmp_size(ext_gmp_type z);

ext_gmp_type ext_gmp_init(void);
void ext_gmp_set_prec (mpf_ptr x, unsigned int prec_in_bits);
void ext_gmp_duplicate_wo_init(ext_gmp_type z1,ext_gmp_type z2);
void ext_gmp_duplicate_w_init(ext_gmp_type z1,ext_gmp_type *z2);



void ext_gmp_add(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,int p);
void ext_gmp_sub(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,int p);
void ext_gmp_mul(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,int p);
void ext_gmp_div(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,int p);
void ext_gmp_abs(ext_gmp_type z1,ext_gmp_type z);
void ext_gmp_truncate(ext_gmp_type z1,ext_gmp_type z);

void ext_gmp_set_z(ext_gmp_type r, int_gmp_type i);

void ext_gmp_add_i(ext_gmp_type z1, int z2, ext_gmp_type z, int p);
void ext_gmp_sub_i(ext_gmp_type z1, int z2, ext_gmp_type z, int p);
void ext_gmp_i_sub(int z1, ext_gmp_type z2, ext_gmp_type z, int p);
void ext_gmp_mul_i(ext_gmp_type z1, int z2, ext_gmp_type z, int p);
void ext_gmp_div_i(ext_gmp_type z1, int z2, ext_gmp_type z, int p);
void ext_gmp_i_div(int z1, ext_gmp_type z2, ext_gmp_type z, int p);

void ext_gmp_writee(ext_gmp_type z,int w);
char* ext_gmp_swritee(ext_gmp_type z, int w);
void ext_gmp_write(ext_gmp_type z,int w);

void ext_gmp_duplicate_w_init(ext_gmp_type z1,ext_gmp_type *z2);
void ext_gmp_duplicate_wo_init(ext_gmp_type z1,ext_gmp_type z2);

void ext_gmp_sqrt(ext_gmp_type z1,ext_gmp_type z,int p);
void ext_gmp_shift(ext_gmp_type z1,ext_gmp_type z,int p);



inline void ext_gmp_set_prec (mpf_ptr x, unsigned int new_prec_in_bits)
{
//	simplified copy of gmp_set_prec
  mp_size_t  old_prec, new_prec, new_prec_plus1;
  mp_size_t  size, sign;
  mp_ptr     xp;

  new_prec = __GMPF_BITS_TO_PREC (new_prec_in_bits);
  old_prec = PREC(x);

  /* do nothing if already the right precision */
  if (new_prec == old_prec)
    return;

  PREC(x) = new_prec;
  new_prec_plus1 = new_prec + 1;

  /* retain most significant limbs */
  sign = SIZ(x);
  size = ABS (sign);
  xp = PTR(x);
  if (size > new_prec_plus1)
    {
      SIZ(x) = (sign >= 0 ? new_prec_plus1 : -new_prec_plus1);
//	removed copying of the values, as we don't need it here:
//	MPN_COPY_INCR (xp, xp + size - new_prec_plus1, new_prec_plus1);
    }

  PTR(x) = __GMP_REALLOCATE_FUNC_LIMBS (xp, old_prec+1, new_prec_plus1);
}


inline ext_gmp_type ext_gmp_init()
{
  ext_gmp_var_count +=1;
  if (gmp_FreeVarCount>0) {
    gmp_FreeVarCount -=1;
    return gmp_FreeVars[gmp_FreeVarCount];
  } else {
  ext_gmp_type z = (ext_gmp_type) malloc(sizeof (__mpf_struct));
  mpf_init(z);
  return z;
  }
}


inline void ext_gmp_duplicate_wo_init(ext_gmp_type z1,ext_gmp_type z2)
{
  unsigned int q1=GMP_PREC(z1);
  if (unlikely(GMP_PREC(z2)< q1)) GMP_SET_PREC(z2,q1);
  mpf_set(z2,z1);
}

inline void ext_gmp_duplicate_w_init(ext_gmp_type z1,ext_gmp_type *z2)
{
  unsigned int q1=GMP_PREC(z1);
  *z2=ext_gmp_init();
  if (unlikely(GMP_PREC(*z2)< q1)) GMP_SET_PREC(*z2,q1);
  mpf_set(*z2,z1);
}

/******************************************************************/

inline int ext_gmp_size(ext_gmp_type z)
{
  unsigned int r;
  if (unlikely(z->_mp_size==0)) return GMP_min;
  count_leading_zeros(r,(z->_mp_d)[abs(z->_mp_size)-1]);
  return (z->_mp_exp) *limb_size_in_bit - r;
}

inline void ext_gmp_add(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,int p)
{ int q,s1,s2;
  s1=ext_gmp_size(z1);
  s2=ext_gmp_size(z2);
  q=MAX_OF(s1,s2)-p+10;
  q=MAX_OF(q,10);
  GMP_SET_PREC(z,q);
  mpf_add(z,z1,z2);
  return;
}






inline void ext_gmp_add_i(ext_gmp_type z1,int n,ext_gmp_type z,int p)
{ 
	int q,s1,s2=limb_size_in_bit;
  	s1=ext_gmp_size(z1);
	q=MAX_OF(s1,s2)-p+10;
	q=MAX_OF(q,10);
	GMP_SET_PREC(z,q);
	if(n<0)mpf_sub_ui(z,z1,-n);
		else mpf_add_ui(z,z1,n);
	return;
}


inline void ext_gmp_sub(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,int p)
{ int q,s1,s2;
  s1=ext_gmp_size(z1);
  s2=ext_gmp_size(z2);
  q=MAX_OF(s1,s2)-p+10;
  q=MAX_OF(q,10);
  GMP_SET_PREC(z,q);
  mpf_sub(z,z1,z2);
  return;
}


inline void ext_gmp_sub_i(ext_gmp_type z1,int z2,ext_gmp_type z,int p)
{ 
	int q,s1,s2=limb_size_in_bit;
  	s1=ext_gmp_size(z1);
	q=MAX_OF(s1,s2)-p+10;
	q=MAX_OF(q,10);
	GMP_SET_PREC(z,q);
  	if(z2<0)mpf_add_ui(z,z1,-z2); else mpf_sub_ui(z,z1,z2);
  	return;
}

inline void ext_gmp_i_sub(int z1,ext_gmp_type z2,ext_gmp_type z,int p)
{ 
	int q,s1,s2=limb_size_in_bit;
  	s1=ext_gmp_size(z2);
	q=MAX_OF(s1,s2)-p+10;
	q=MAX_OF(q,10);
  	GMP_SET_PREC(z,q);
	if(z1<0){mpf_add_ui(z,z2,-z1);mpf_neg(z,z);} else mpf_ui_sub(z,z1,z2);
	return;
}



inline void ext_gmp_mul(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,int p)
{ int q,s1,s2;
  s1=ext_gmp_size(z1);
  s2=ext_gmp_size(z2);
  q=s1+s2-p+10;
  q=MAX_OF(q,10);
  GMP_SET_PREC(z,q);
  mpf_mul(z,z1,z2);
  return;
}

inline void ext_gmp_mul_i(ext_gmp_type z1, int z2, ext_gmp_type z, int p)
{
int q,s1,zz,s2=limb_size_in_bit;
s1=ext_gmp_size(z1);
q=s1+s2-p+10;
q=MAX_OF(q,10);
GMP_SET_PREC(z,q);
if (z2<0) zz=-z2; else zz=z2;
mpf_mul_ui(z,z1,zz);
if (z2<0)mpf_neg(z,z);
return;
}

inline void ext_gmp_div(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,int p)
{ int q,s1,s2;
  s1=ext_gmp_size(z1);
  s2=ext_gmp_size(z2);
  q=s1-s2-p+10;
  q=MAX_OF(q,10);
  GMP_SET_PREC(z,q);
  mpf_div(z,z1,z2);
  return;
}

inline void ext_gmp_div_i(ext_gmp_type z1,int z2,ext_gmp_type z,int p)
{ 
	int q,s1,zz,s2=1;
	s1=ext_gmp_size(z1);
	q=s1-s2-p+10;
	q=MAX_OF(q,10);
	GMP_SET_PREC(z,q);
	if(z2<0)zz=-z2;else zz=z2;
	mpf_div_ui(z,z1,zz);
	if(z2<0)mpf_neg(z,z);
  	return;
}

inline void ext_gmp_i_div(int z1,ext_gmp_type z2,ext_gmp_type z,int p)
{ 
	int q,s1=limb_size_in_bit,zz,s2;
	s2=ext_gmp_size(z2);
	q=s1-s2-p+10;
	q=MAX_OF(q,10);
	GMP_SET_PREC(z,q);
	if(z1<0)zz=-z1;else zz=z1;
	mpf_ui_div(z,zz,z2);
	if(z1<0)mpf_neg(z,z);
	return;
}




inline void ext_gmp_abs(ext_gmp_type z1,ext_gmp_type z)
{
  int q1=GMP_PREC(z1);
  GMP_SET_PREC(z,q1);
  mpf_abs(z,z1);
  return;
}

inline void ext_gmp_truncate(ext_gmp_type z1,ext_gmp_type z)
{
  int q1=GMP_PREC(z1);
  GMP_SET_PREC(z,q1);
  mpf_trunc(z,z1);
  return;
}


inline void ext_gmp_writee(ext_gmp_type z,int w)
{ char s[w];
  long e; 
  int i;
  int flag=0;
  mpf_get_str (s, &e, 10, w-7,z);
  for (i=0;i<w-7;i++) {
    if ( s[i]==0 ) flag=1 ;
    if ( flag ) s[i]='0';
  } 
  if (s[0]=='-') {s[w-7]=0;printf("-.%sE%+05ld",&s[1],e);}
  else {s[w-8]=0;printf("+.%sE%+05ld",&s[0],e);};
}

inline char* ext_gmp_swritee(ext_gmp_type z,int w)
//ASSERT: w must be so small that 3.5*w still has no overflow ! 
{
// At the moment, we frist vcreate the result in r, then copy it to s.
// This could be simplified.
  char *s=(char *)malloc(w+1);
  char r[w];
  long e, e_abs,r_len; 
  int i;
  int flag=0;
  int sgn =  mpf_sgn (z);
  /* 
  The conversion routine mpf_get_str does not only depend on the 
  numerical value of the result, but also on the assigned precision:
    "No more digits than can be accurately represented
     by OP are ever generated."
  In consequence, we need to `add' precision to get access to the
  numerical value of z by copying its value to z_copy !     
  */
  mpf_t z_copy;
  long bin_w=(long) (w*3.5);
  mpf_init2(z_copy,bin_w);
  mpf_abs(z_copy,z);

  mpf_get_str(r, &e, 10, w-4,z_copy);

  mpf_clear(z_copy);
  /* 
  Next problem: mpf_get_str discards trailing zeroes,
  so we have to add them again...
  */  
  for (i=0;i<w-5;i++) {
    if ( r[i]==0 ) flag=1 ;
    if ( flag ) r[i]='0';
  }

  /*
  we want to have an error of less than 0.51ulp,
  so we took two additional digits giving an error of at most 0.01ulp
  (concerning the intended precision).
  The position of these digits is (w-8,w-7) for the 
  usual 4-decimal exponents and decreases for the larger cases.
  If the first of the two additional digits is 5 or greater,
  we round upwards (giving at most additional 0.5ulp)
  */ 


  r_len=e; if (r_len<0) r_len=-r_len;
  if (r_len<10000){r_len=4;}
  else if (r_len<100000){r_len=5;}
  else if (r_len<1000000){r_len=6;}
  else if (r_len<10000000){r_len=7;}
  else if (r_len<100000000){r_len=8;}
  else if (r_len<1000000000){r_len=9;}
  else {r_len=10;};

  /*
  r_len is the number of decimal places of the exponent prior to the rounding
  */


  if ( w < r_len+5) w=r_len+5;
  /*
  Produce at least one significant digit...
  */


  if ( r[w-r_len-4]>='5' ){
    for (i=w-r_len-5;i>=0;i--){
      if ( r[i] < '9' ) {
          r[i]++;
          break;
      }
      r[i]='0';
    }
    /* 
    In the rare case that we switch from 0.999...999 (length n) to 1.000...000 (length n+1), 
    we have to correct the exponent and the mantissa, but changing the mantissa is simple:
    (1) If the (decimal) length of the exponent does not change, result is 0.100...000 (length n) 
    (2) If the (decimal) length of the exponent increases, result is 0.100...000 (length n-1) 
    (3) If the (decimal) length of the exponent decreases, result is 0.100...000 (length n+1)
    So we simply set the last decimal to '0'. The 0.51ulp accuracy is true in all three cases!
    */ 
    if (i<0){
      r[0]='1';
      r[w-r_len-4]='0';
      e++;
    }
  }
  r[w-r_len-3] = 0;
  if (sgn== -1) {sprintf(s,"-.%s",&r[0]);}
  else {sprintf(s,"+.%s",&r[0]);};
  e_abs=e; if (e_abs<0) e_abs=-e_abs;
  if (e_abs<10000){e_abs=4;}
  else if (e_abs<100000){e_abs=5;}
  else if (e_abs<1000000){e_abs=6;}
  else if (e_abs<10000000){e_abs=7;}
  else if (e_abs<100000000){e_abs=8;}
  else if (e_abs<1000000000){e_abs=9;}
  else {e_abs=10;};
  if (e_abs > w-2){sprintf(&s[w-6],"E*****");
  } else {
    if (e_abs==4){sprintf(&s[w-6],"E%+05ld",e);}
    else if (e_abs==5){sprintf(&s[w-7],"E%+06ld",e);}
    else if (e_abs==6){sprintf(&s[w-8],"E%+07ld",e);}
    else if (e_abs==7){sprintf(&s[w-9],"E%+08ld",e);}
    else if (e_abs==8){sprintf(&s[w-10],"E%+09ld",e);}
    else if (e_abs==9){sprintf(&s[w-11],"E%+010ld",e);}
    else {sprintf(&s[w-12],"E%+011ld",e);};
  }
  return s;
}

inline void ext_gmp_write(ext_gmp_type z,int w)
{ char s[w];
  long e; 
  int i;
  int flag=0;
  mpf_get_str (s, &e, 10, w-7,z);
  for (i=0;i<w-7;i++) { if ( s[i]==0 ) flag=1 ;
    if ( flag ) s[i]='0';
  } 
  if (s[0]=='-') {s[w-7]=0;printf("-.%sE%+05ld",&s[1],e);}
  else {s[w-8]=0;printf("+.%sE%+05ld",&s[0],e);};
  if (z->_mp_size != 0) {
     printf("\n");
     for (i=abs(z->_mp_size-1);i>=0; i-=1) 
       printf("%lu ",z->_mp_d[i]);
  }
  printf("\nExponent %ld, Limbs: %d, Bits: %uld\n",
     z->_mp_exp,z->_mp_size,GMP_PREC(z));
}

inline void ext_gmp_sqrt(ext_gmp_type z1,ext_gmp_type z,int p)
{ int q,s1;
  s1=ext_gmp_size(z1);
  q=s1/2-p+10;
  q=MAX_OF(q,10);
  GMP_SET_PREC(z,q);
  mpf_sqrt(z,z1);
  return;
}

inline void ext_gmp_shift(ext_gmp_type z1,ext_gmp_type z,int n)
{
  GMP_SET_PREC(z,GMP_PREC(z1));
  if ( n>=  0 ) mpf_mul_2exp(z,z1,n);
  else mpf_div_2exp(z,z1,-n);
  return;
}
inline void ext_gmp_set_z(ext_gmp_type r,int_gmp_type i)
{
  GMP_SET_PREC(r,mpz_sizeinbase(i,2));
  mpf_set_z(r,i);
}

#ifdef __cplusplus
}
#endif

#endif /* ifndef GMP_ext*/

