/*

gmp_extension.cc -- extension to the GMP float functions for the iRRAM library
 
Copyright (C) 2002-2006 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/

#include "iRRAM_core.h"
#include "gmp_extension.h"

using namespace iRRAM;

typedef  DYADIC (DYADIC_function) (DYADIC,int);

inline void mpf_wrapper(DYADIC_function f, mpf_ptr r, mpf_srcptr u,int p){
  DYADIC arg,res;
  mpf_set_prec(arg.value,mpf_get_prec(u));
  mpf_set(arg.value,u);
  res = iRRAM_exec(f,arg,p);
  mpf_set_prec(r,mpf_get_prec(res.value));
  mpf_set(r,res.value);
}

DYADIC compute_cos(DYADIC x, int p) { return approx(cos(REAL(x)),p); }
void mpf_ext_cos (mpf_ptr r, mpf_srcptr u,int p){mpf_wrapper(compute_cos,r,u,p);};

DYADIC compute_sin(DYADIC x, int p) { return approx(sin(REAL(x)),p); }
void mpf_ext_sin (mpf_ptr r, mpf_srcptr u,int p){mpf_wrapper(compute_sin,r,u,p);};

DYADIC compute_exp(DYADIC x, int p) { return approx(exp(REAL(x)),p); }
void mpf_ext_exp (mpf_ptr r, mpf_srcptr u,int p){mpf_wrapper(compute_exp,r,u,p);};

DYADIC compute_log(DYADIC x, int p) { return approx(log(REAL(x)),p); }
void mpf_ext_log (mpf_ptr r, mpf_srcptr u,int p){mpf_wrapper(compute_log,r,u,p);};

DYADIC compute_tan(DYADIC x, int p) { return approx(tan(REAL(x)),p); }
void mpf_ext_tan (mpf_ptr r, mpf_srcptr u,int p){mpf_wrapper(compute_tan,r,u,p);};



